/*
 * Copyright 2009 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "globus_common.h"
#include "globus_gsi_authz.h"
#include "globus_gss_assist.h"
#include "globus_i_gsi_authz_esgsaml_callout.h"
#include "openssl/ssl.h"
#include "openssl/err.h"
#include "openssl/x509.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include "esg_saml.h"
#include "libglobus_adq/libglobus_adq.h"


#ifndef BUILD_DEBUG
#define BUILD_DEBUG
#endif

#define MAX_URI_LEN 512

#define DEFAULT_AUTH_SERVICE (char *)"http://localhost:8080/SamlAuthzProject/saml/soap/secure/authorizationService.htm"

extern int errno;

int      globus_i_gsi_authz_esgsaml_callout_debug_level   = 0;
FILE *   globus_i_gsi_authz_esgsaml_callout_debug_fstream = 0;

typedef struct
{
    char *auth_service;
    char *saml_assertion;
    char *authz_assert;
    char *openid;
    int got_saml_assertion_over_control_channel;
    authz_decision_stmt_obj_t authz_decision_stmt;
} authz_esgsaml_state_t;

static char *get_error_msg(int ret)
{
    return globus_object_printable_to_string(
        globus_error_get((globus_result_t)ret));
}

/*
 * ap is:
 *		void * authz_system_state;
 */
globus_result_t
authz_esgsaml_system_init_callout(
    va_list                             ap)
{
    void **authz_system_state;
    
    globus_result_t                 result = GLOBUS_SUCCESS;
    static char *                   _function_name_ =
	"authz_esgsaml_system_init_callout";

    struct stat buf;
    FILE *fconfig = NULL; 
    char *ptr = NULL, *dest = NULL;
    authz_esgsaml_state_t *esgsaml_state = NULL;
    char line[512] = {0}, key[128] = {0}, value[128] = {0};
    char *config_file = "/etc/grid-security/esgsaml_auth.conf";
    char *tmp_string = 0;

    tmp_string = globus_module_getenv("GLOBUS_GSI_AUTHZ_DEBUG_LEVEL");
    if (tmp_string != GLOBUS_NULL)
    {
        globus_i_gsi_authz_esgsaml_callout_debug_level = atoi(tmp_string);
        
        if(globus_i_gsi_authz_esgsaml_callout_debug_level < 0)
        {
            globus_i_gsi_authz_esgsaml_callout_debug_level = 0;
        }
    }

    tmp_string = globus_module_getenv("GLOBUS_GSI_AUTHZ_DEBUG_FILE");
    if (tmp_string != GLOBUS_NULL)
    {
        globus_i_gsi_authz_esgsaml_callout_debug_fstream = fopen(tmp_string, "a");
    }

    if (globus_i_gsi_authz_esgsaml_callout_debug_fstream == 0)
    {
        /* if the env. var. isn't set (or the fopen failed), use stderr */
        globus_i_gsi_authz_esgsaml_callout_debug_fstream = stderr;
    }

    authz_system_state = va_arg(ap, void **);

    if ((esgsaml_state = globus_libc_calloc(1, sizeof(authz_esgsaml_state_t))) == 0)
    {
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
            "Failed to allocate esgsaml state! (ret=%d, errno=%d)",
            result, errno);
        goto end;
    }

    esgsaml_state->auth_service = NULL;
    esgsaml_state->saml_assertion = NULL;
    esgsaml_state->got_saml_assertion_over_control_channel = 0;
    memset(&esgsaml_state->authz_decision_stmt, 0, sizeof(authz_decision_stmt_obj_t));

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
	GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
	"in %s, system state is %p\n",
	_function_name_, authz_system_state);

    memset(&buf, 0, sizeof(struct stat));
    if (stat(config_file, &buf) != 0)
    {
        config_file = getenv("ESGSAML_AUTH_CONFIG_FILE");
        if (stat(config_file, &buf) != 0)
        {
            fprintf(stderr, "No configuration file found.  Using defaults.\n");
            goto end;
        }
    }

    fconfig = fopen(config_file, "r");
    if (fconfig)
    {
        while (fgets(line, 512, fconfig) != NULL)
        {
            if (line[0] == '#')
            {
                continue;
            }
            else
            {
                memset(key, 0, 128);
                memset(value, 0, 128);
                    
                for(ptr = line, dest = key; *ptr != '\0'; ptr++)
                {
                    if (*ptr == '=')
                    {
                        dest = value;
                    }
                    else if ((*ptr == ' ') || (*ptr == '\t') || (*ptr == '\r') || (*ptr == '\n'))
                    {
                        continue;
                    }
                    else
                    {
                        *dest++ = *ptr;
                    }
                }
                if (strcasecmp(key, "authservice") == 0)
                {
                    esgsaml_state->auth_service = strdup(value);
                }
/*                 else */
/*                 { */
/*                     fprintf(stderr, "Unrecognized key found: %s\n", key); */
/*                 } */
            }
        }
        fclose(fconfig);
    }

  end:

    if (strlen(esgsaml_state->auth_service) == 0)
    {
        strcpy(esgsaml_state->auth_service, DEFAULT_AUTH_SERVICE);
    }

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
        "Using configuration option AUTHSERVICE=%s\n",
        esgsaml_state->auth_service);

    if (result == GLOBUS_SUCCESS)
    {
        *authz_system_state = (void *)esgsaml_state;
    }
    else
    {
        if (esgsaml_state)
        {
            free(esgsaml_state);
            esgsaml_state = NULL;
        }
    }
    return result;
}

globus_result_t
authz_esgsaml_system_destroy_callout(
    va_list                             ap)
{
    void ** authz_system_state;
    
    globus_result_t                 result = GLOBUS_SUCCESS;
    static char *                   _function_name_ =
	"authz_esgsaml_system_destroy_callout";

    authz_system_state = va_arg(ap, void **);

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
	GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
	"in %s, system state is %p\n", _function_name_,
	 authz_system_state);

#ifdef BUILD_DEBUG
    if (globus_i_gsi_authz_esgsaml_callout_debug_fstream &&
	(globus_i_gsi_authz_esgsaml_callout_debug_fstream != stderr))
    {
        fflush(globus_i_gsi_authz_esgsaml_callout_debug_fstream);
	fclose(globus_i_gsi_authz_esgsaml_callout_debug_fstream);
	globus_i_gsi_authz_esgsaml_callout_debug_fstream = stderr;
    }
#endif /* BUILD_DEBUG */

    return result;
}

globus_result_t
authz_esgsaml_handle_init_callout(
    va_list                             ap)
{
    char * service_name;
    gss_ctx_id_t context;
    globus_gsi_authz_cb_t callback;
    void * callback_arg;
    void * authz_system_state;
    OM_uint32 maj_stat, min_stat;
    gss_name_t src_name, targ_name;
    gss_buffer_desc sname, tname;
    gss_OID name_type;
    globus_gsi_authz_handle_t *handle;
    globus_result_t result = !GLOBUS_SUCCESS;
    static char *_function_name_ =
        "authz_esgsaml_handle_init_callout";
    authz_esgsaml_state_t *esgsaml_state = NULL;
    char *DN = (char *)0, *start = (char *)0, *end = (char *)0;

    handle = va_arg(ap, globus_gsi_authz_handle_t *);
    service_name = va_arg(ap, char *);
    context = va_arg(ap, gss_ctx_id_t);
    callback = va_arg(ap,  globus_gsi_authz_cb_t);
    callback_arg = va_arg(ap, void *);
    authz_system_state = va_arg(ap, void *);

    esgsaml_state = (authz_esgsaml_state_t *)authz_system_state;

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF5(
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
        "in %s\n\tservice name is %s\n\tcontext is %x, system state is %p\n",
        _function_name_, service_name, (unsigned)context, authz_system_state);

    /* pull OpenID from user's certificate DN/Subject */
    maj_stat = gss_inquire_context(&min_stat, context,
                                   &src_name, &targ_name, NULL,
                                   NULL, NULL, NULL, NULL);
    if (maj_stat != GSS_S_COMPLETE)
    {
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
            "gss_inquire_content failed: %s\n", get_error_msg(min_stat));
        goto end;
    }

    maj_stat = gss_display_name(&min_stat, src_name, &sname,
                                &name_type);
    if (maj_stat != GSS_S_COMPLETE)
    {
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
            "gss_display_name(source) failed: %s\n", get_error_msg(min_stat));
        goto end;
    }
    maj_stat = gss_display_name(&min_stat, targ_name, &tname,
                                (gss_OID *) NULL);
    if (maj_stat != GSS_S_COMPLETE)
    {
        /* if we have the sname, we can ignore if the tname fails */
        if (sname.value == NULL)
        {
            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
                GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                "gss_display_name(target) failed: %s\n", get_error_msg(min_stat));
            goto end;
        }
    }

    DN = ((sname.value != NULL) ? strdup(sname.value) : strdup(tname.value));

    (void) gss_release_name(&min_stat, &src_name);
    (void) gss_release_name(&min_stat, &targ_name);
    (void) gss_release_buffer(&min_stat, &sname);
    (void) gss_release_buffer(&min_stat, &tname);

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
        "Got user DN: %s\n", (char *)DN);

    /* parse out OPENID from user's DN here */
    start = DN;
    end = (DN + strlen(DN));
    for(start = DN; start != end; start++)
    {
        if ((*start == 'h') && ((int)(end - start) > 6))
        {
            if (strncmp(start, "http", 4) == 0)
            {
                esgsaml_state->openid = strdup(start);

                GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
                    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                    "Found user's OpenID: %s\n", (char *)esgsaml_state->openid);
                break;
            }
        }
    }

    free(DN);
    *handle = globus_libc_calloc(1, sizeof(globus_gsi_authz_cb_t));

    if (esgsaml_state->openid != NULL)
    {
        result = GLOBUS_SUCCESS;
    }
    else
    {
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
            "User's OpenID could not be found.  Cannot proceed with authorization.", "\n");
    }

  end:
    if (result == GLOBUS_SUCCESS)
    {
        callback(callback_arg, callback_arg, result);
    }

    return result;
}


globus_result_t
authz_esgsaml_authorize_async_callout(
    va_list                             ap)
{
    int ret = -1;
    globus_gsi_authz_handle_t handle;
    char * action;
    char * object;
    globus_gsi_authz_cb_t callback;
    void * callback_arg;
    void * authz_system_state;

    globus_result_t                 result = !GLOBUS_SUCCESS;
    static char *                   _function_name_ =
        "authz_esgsaml_authorize_async_callout";
    authz_esgsaml_state_t *esgsaml_state = NULL;

    handle = va_arg(ap, globus_gsi_authz_handle_t);
    action = va_arg(ap, char *);
    object = va_arg(ap, char *);
    callback = va_arg(ap,  globus_gsi_authz_cb_t);
    callback_arg = va_arg(ap, void *);
    authz_system_state = va_arg(ap, void *);
    esgsaml_state = (authz_esgsaml_state_t *)authz_system_state;

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF5(
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
        "in %s, action is %s, object is %s, system state is %p\n",
        _function_name_, action, object, authz_system_state);

    if (action == NULL)
    {
        return result;
    }

    if (strcmp(action, "authz_assert") == 0)
    {
        globus_assert(esgsaml_state->saml_assertion == NULL);

        if (object != NULL)
        {
            char *ptr = object;
            char *endptr = (object + strlen(object));
            while(ptr < endptr)
            {
                if (*ptr == '<')
                {
                    /* store the saml assertion passed via authz_assert for later */
                    esgsaml_state->saml_assertion = strdup(ptr);
                    break;
                }
                ptr++;
            }

            if (ptr == endptr)
            {
                char *gsi_start = strstr(object, "gsiftp");
                if (gsi_start)
                {
                    char *gsi_end = (gsi_start + strlen(gsi_start) - 1);
                    char *cur = gsi_start;

                    // if the gsiftp URL was passed via control channel, strip out the port
                    while(cur < gsi_end)
                    {
                        if (*cur == '?')
                        {
                            *cur = '\0';
                            break;
                        }
                        cur++;
                    }

                    // attempt to trim trailing carriage return if any
                    if (gsi_start && (gsi_start[strlen(gsi_start)-1] == '\n'))
                    {
                        gsi_start[strlen(gsi_start)-1] == '\0';
                    }

                    esgsaml_state->authz_assert = strdup(gsi_start);

                    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
                        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                        "Got authz assert gsiftp URL: %s\n", esgsaml_state->authz_assert);

                    esgsaml_state->got_saml_assertion_over_control_channel = 0;
                    result = GLOBUS_SUCCESS;
                }
                else
                {
                    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
                        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                        "Failed to parse SAML assertion:\n%s\n",
                        esgsaml_state->saml_assertion);

                    goto end;
                }
            }
            else
            {
                // otherwise try to parse it as a SAML assertion
                ret = parse_saml_assertion(esgsaml_state->saml_assertion,
                                           &esgsaml_state->authz_decision_stmt);
                if (ret != 0)
                {
                    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
                        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                        "Failed to parse SAML assertion:\n%s\n",
                        esgsaml_state->saml_assertion);

                    goto end;
                }

                esgsaml_state->got_saml_assertion_over_control_channel = 1;
                result = GLOBUS_SUCCESS;
            }
        }
    }
    else if (strcmp(action, "read") == 0)
    {
        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
            "Got Action \"%s\" with object \"%s\"\n", action, object);

        /*
          if we didn't receive an assertion over the control channel,
          query the auth service directly for this file
        */
        if (esgsaml_state->got_saml_assertion_over_control_channel == 0)
        {
            char *query_file = object;
            authz_decision_t ads;
            memset(&ads, 0, sizeof(ads));

            // if we were passed a gsiftp URL via the authz assert, we
            // need to use it b/c it has port information
            if (esgsaml_state->authz_assert)
            {
                // but make sure that the file matches something in
                // the passed in gsiftp URL
                char *obj_file = (object + strlen(object) - 1);
                while(obj_file > object)
                {
                    if (*obj_file == '/')
                    {
                        obj_file++;
                        break;
                    }
                    obj_file--;
                }

                if (strstr(esgsaml_state->authz_assert, obj_file))
                {
                    query_file = esgsaml_state->authz_assert;
                }
            }

            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF5(
                GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                "Calling out to auth service %s to retrieve SAML Assertion\n\tfor "
                "identity %s, file %s, and action %s\n", esgsaml_state->auth_service,
                esgsaml_state->openid, query_file, action);

            ret = get_authorization_decision(
                esgsaml_state->auth_service, esgsaml_state->openid,
                query_file, action, &ads);

            if (ret != 0)
            {
                GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
                    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                    "Failed to retrieve SAML assertion response for %s\n",
                    esgsaml_state->openid);
                goto end;
            }
            else if ((ret == 0) && (ads.decision != ADQ_PERMIT))
            {
                GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2(
                    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
                    "Retrieved SAML assertion response for %s, but response is not \"PERMIT\"\n",
                    esgsaml_state->openid);
                goto end;
            }
            result = ((ret == 0) ? GLOBUS_SUCCESS : !GLOBUS_SUCCESS);
        }
        else
        {
            /* this part is not yet supported ... */

/*             ret = parse_saml_assertion(esgsaml_state->saml_assertion, */
/*                                        &esgsaml_state->authz_decision_stmt); */

/*             if (ret != 0) */
/*             { */
/*                 GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2( */
/*                     GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE, */
/*                     "Failed to parse SAML assertion:\n%s\n", */
/*                     esgsaml_state->saml_assertion); */

/*                 goto end; */
/*             } */

/*             GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2( */
/*                 GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE, */
/*                 "Parsed %d AuthzDecisionStatements.\n", */
/*                 esgsaml_state->authz_decision_stmt.cur_index); */

/*             ret = assertion_contains_object(&esgsaml_state->authz_decision_stmt, object); */

/*             GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF2( */
/*                 GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE, */
/*                 "Verdict of object transfer: %s\n", */
/*                 ((ret == 0) ? "ALLOW" : "DENY")); */

/*             result = ((ret == 0) ? GLOBUS_SUCCESS : !GLOBUS_SUCCESS); */
        }
    }
    else
    {
        /*
          some other operation occured (e.g. "lookup").  let's return
          the value of whether or not the transfer went through each
          time.  so if the transfer already happened, we can return
          GLOBUS_SUCCESS, otherwise a failure.
        */
/*         if (esgsaml_state->esgsaml != NULL) */
/*         { */
/*             result = esgsaml_state->file_transferred; */
/*         } */

        result = ((action != NULL) ? GLOBUS_SUCCESS : !GLOBUS_SUCCESS);

        GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
            GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
            "Got operation \"%s\" and returning: %s\n", action,
            ((result == GLOBUS_SUCCESS) ? "ALLOW" : "DENY"));
    }

  end:

    if (result == GLOBUS_SUCCESS)
    {
        callback(callback_arg, handle, result);
    }
    return result;
}

int
authz_esgsaml_cancel_callout(
    va_list                             ap)
{
    globus_gsi_authz_handle_t           handle;
    void * 				authz_system_state;

  int                             	result = (int) GLOBUS_SUCCESS;
  static char *                   	_function_name_ =
    "authz_esgsaml_cancel_callout";

    handle = va_arg(ap, globus_gsi_authz_handle_t);
    authz_system_state = va_arg(ap, void *);

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
	GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
	"in %s, system state is %p\n",
	_function_name_, authz_system_state);

    return result;
}

int
authz_esgsaml_handle_destroy_callout(
    va_list                             ap)
{
    globus_gsi_authz_handle_t 		handle;
    void **				authz_system_state;
    
    int                             	result = (int) GLOBUS_SUCCESS;
    static char *                   	_function_name_ =
	"authz_esgsaml_handle_destroy_callout";
    globus_gsi_authz_cb_t		callback;
    void *				callback_arg;
    authz_esgsaml_state_t *esgsaml_state = NULL;
    int i = 0;
    
    handle = va_arg(ap, globus_gsi_authz_handle_t);
    callback = va_arg(ap, globus_gsi_authz_cb_t);
    callback_arg = va_arg(ap, void *);
    authz_system_state = va_arg(ap, void **);

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
	GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
	"in %s, system state is %p\n",
	_function_name_, authz_system_state);

    esgsaml_state = (authz_esgsaml_state_t *)(*authz_system_state);
    if (esgsaml_state != NULL)
    {
        for(i = 0; i < esgsaml_state->authz_decision_stmt.cur_index; i++)
        {
            if (esgsaml_state->authz_decision_stmt.action[i])
            {
                free(esgsaml_state->authz_decision_stmt.action[i]);
                esgsaml_state->authz_decision_stmt.action[i] = NULL;
            }
            if (esgsaml_state->authz_decision_stmt.resource[i])
            {
                free(esgsaml_state->authz_decision_stmt.resource[i]);
                esgsaml_state->authz_decision_stmt.resource[i] = NULL;
            }
            if (esgsaml_state->authz_decision_stmt.decision[i])
            {
                free(esgsaml_state->authz_decision_stmt.decision[i]);
                esgsaml_state->authz_decision_stmt.decision[i] = NULL;
            }
        }

        if (esgsaml_state->auth_service)
        {
            free(esgsaml_state->auth_service);
            esgsaml_state->auth_service = NULL;
        }
        if (esgsaml_state->saml_assertion)
        {
            free(esgsaml_state->saml_assertion);
            esgsaml_state->saml_assertion = NULL;
        }
        if (esgsaml_state->authz_assert)
        {
            free(esgsaml_state->authz_assert);
            esgsaml_state->authz_assert = NULL;
        }

        globus_libc_free(esgsaml_state);
        esgsaml_state = NULL;
        *authz_system_state = NULL;

        if (handle != NULL)
        {
            globus_libc_free(handle);
        }
        handle = NULL;
    }

    callback(callback_arg, handle, result);
    return result;
}

int
globus_gsi_authz_esgsaml_get_authorization_identity_callout(
    va_list                             ap)
{
    globus_gsi_authz_handle_t		handle;
    char **				identity_ptr;
    globus_gsi_authz_cb_t		callback;
    void *				callback_arg;
    void * 				authz_system_state;
    
    globus_result_t                    	result = GLOBUS_SUCCESS;
    static char *                   	_function_name_ =
	"globus_gsi_authz_esgsaml_get_authorization_identity_callout";

    handle = va_arg(ap, globus_gsi_authz_handle_t);
    identity_ptr = va_arg(ap, char **);
    callback = va_arg(ap, globus_gsi_authz_cb_t);
    callback_arg = va_arg(ap, void *);
    authz_system_state = va_arg(ap, void *);

    GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_FPRINTF3(
	GLOBUS_I_GSI_AUTHZ_ESGSAML_CALLOUT_DEBUG_TRACE,
	"in %s, system state is %p\n",
	_function_name_, authz_system_state);

    return((int) result);
}
